public class CpmDpb {
	public int spt; // 128-byte sectors
	public int bsh;
	public int blm;
	public int exm;
	public int dsm; // total number of blocks, - 1
	public int drm; // total number of dir entries, - 1
	public int al0; // bit 15 is for block 0!
	public int cks;
	public int off; // dir offset, tracks (spt)
	public int poff; // not CP/M standard, partition offset (records)
	public int pssz; // ('') physical sector size
	public int plat; // ('') physical sector latency (sectrn used by CP/M)

	public CpmDpb() {}

	public CpmDpb(int[] vals) {
		spt = vals[0];
		bsh = vals[1];
		blm = vals[2];
		exm = vals[3];
		dsm = vals[4];
		drm = vals[5];
		al0 = vals[6];
		cks = vals[7];
		off = vals[8];
		if (vals.length > 9) {
			poff = vals[9];
		} else {
			poff = 0;
		}
		if (vals.length > 10) {
			pssz = vals[10];
		} else {
			pssz = 0;
		}
		if (vals.length > 11) {
			plat = vals[11];
		} else {
			plat = 0;
		}
	}

	public boolean hasReserved() {
		int bs = (1 << bsh);	// block size, records
		int db = (((drm + 1 + 3) / 4) + bs - 1) / bs;	// num blocks for dir
		int rb = Integer.bitCount(al0);	// resv blocks
		return (db < rb);
	}

	public int numReserved() {
		int bs = (1 << bsh);	// block size, records
		int db = (((drm + 1 + 3) / 4) + bs - 1) / bs;	// num blocks for dir
		int rb = Integer.bitCount(al0);	// resv blocks
		return (rb - db);
	}

	public int firstReserved() {
		int bs = (1 << bsh);	// block size, records
		int db = (((drm + 1 + 3) / 4) + bs - 1) / bs;	// num blocks for dir
		return db;	// also, first block after dir...
	}
}
