// Copyright (c) 2018 Douglas Miller <durgadas311@gmail.com>

import java.io.*;

public class memdump {
	public static void main(String[] args) {
		File core = null;
		DebugMemory mem;
		boolean usage = false;
		boolean dis = false;
		boolean zilog = false;
		int base = -1;
		int top = -1;
		int org = 0;

		for (String arg : args) {
			File f = new File(arg);
			if (f.exists()) {
				core = f;
				continue;
			}
			if (arg.startsWith("dis") || arg.equalsIgnoreCase("dasm")) {
				dis = true;
				continue;
			}
			if (arg.equalsIgnoreCase("zilog")) {
				zilog = true;
				continue;
			}
			if (arg.equalsIgnoreCase("mac80")) {
				zilog = false;
				continue;
			}
			if (arg.startsWith("org=")) {
				org = Integer.valueOf(arg.substring(4), 16);
				continue;
			}
			if (arg.matches("[0-9A-Fa-f]*")) {
				int h = Integer.valueOf(arg, 16);
				if (base < 0) {
					base = h;
				} else if (top < 0) {
					top = h;
				} else {
					System.err.format("extra HEX arg: %s\n", arg);
					usage = true;
				}
				continue;
			}
			System.err.format("Unrecognized arg: \"%s\"\n", arg);
			usage = true;
		}
		if (usage || core == null) {
			System.err.format("Usage: memdump <image> [dis [zilog|mac80]] [org=<org>] [<base> [<top>]]\n");
			System.exit(1);
		}
		if (dis) {
			mem = new DebugMemory(core, org, zilog);
		} else {
			mem = new DebugMemory(core, org);
		}
		if (base < org) { // also if base == -1
			base = org;
		}
		if (top < 0) {
			top = mem.size() + org;
		}
		if (!dis) {
			base &= ~0x0f;
			top = (top + 0x0f) & ~0x0f;
		}
		if (top < base) {
			System.err.format("Impossible range values %05x..%05x\n", base, top);
			System.exit(1);
		}
		for (int adr = base; adr < top;) {
			if (dis) {
				int n = mem.disas(adr);
				adr += n;
			} else {
				mem.dumpLine(adr);
				adr += 16;
			}
		}
	}
}
