// Copyright (c) 2020 Douglas Miller <durgadas311@gmail.com>
import java.io.*;

public interface CpnetSerialProtocol {
	int getRetries();

	void setIO(InputStream in, OutputStream out);

	// Returns true if char indicates start of protocol message
	// (else char is console output - if feature enabled).
	// Not used when console feature disabled.
	boolean protoStart(int ch);

	// Wait indefinitely for a protocol message.
	// Returns 0 on success.
	// 	  -1 on permanent comm failure (do not retry)
	//	   1 on "NAK sent" (checksum error)
	//	   2 on timeout (client gone?)
	//	   3 on protocol error
	int recv(byte[] buf, int off, int len);

	// After protoStart() returns true, this is called
	// to receive the message that was started.
	// 'ch' is the protocol character tested with protoStart().
	// Returns same as recv(byte[] buf, int off, int len).
	int recv(int ch, byte[] buf, int off, int len);

	// Returns 0 on success.
	// 	  -1 on permanent comm failure (do not retry)
	//	   1 on "NAK"
	//	   2 on timeout (client gone?)
	int send(byte[] buf, int off, int len);
}
