// Copyright (c) 2020 Douglas Miller <durgadas311@gmail.com>

import java.io.InputStream;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.util.Properties;

public class EPROM implements Rom {
	protected byte[] rom;
	protected int org;
	protected int msk;

	private InputStream openROM(String rom) throws Exception {
		Exception ex = null;
		try {
			return new FileInputStream(rom);
		} catch (Exception ee) {
			ex = ee;
		}
		try {
			// This returns null if not found!
			InputStream fi = this.getClass().getResourceAsStream(rom);
			if (fi != null) {
				return fi;
			}
		} catch (Exception ee) {
			ex = ee;
		}
		if (ex != null) {
			throw ex;
		}
		throw new FileNotFoundException(rom);
	}

	public int length() { return rom.length; }

	public EPROM(String img, int base) throws Exception {
		org = base;
		InputStream fi = openROM(img);
		msk = fi.available() - 1;
		rom = new byte[msk + 1];
		fi.read(rom);
		fi.close();
	}

	public int read(int addr) {
		return rom[addr & msk] & 0xff;
	}

	public void write(int addr, int val) {}
	public void writeEnable(boolean we) {}

	public String dumpDebug() {
		int x = (rom.length * 8) / 1024;	// TODO: power of 2
		return String.format("27%02d EPROM at %04x\n", x, org);
	}
}
