// Copyright (c) 2016 Douglas Miller <durgadas311@gmail.com>

import java.io.InputStream;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.util.Properties;

public class H89Roms {
	protected Rom mon; // 2K or 4K, or 32K EEPROM, at 0x0000
	protected Rom h17; // always 2K at 0x1800
	protected int romMask;
	protected int monMask;
	protected int romBit = 0;
	protected int gpioMEM1 = 0;
	protected int gpioIO1 = 0;

	public H89Roms(Properties props, GeneralPurposePort gpio, Interruptor intr) {
		InputStream fi = null;
		boolean isH8 = H89.isH8(props);
		boolean isZ80 = H89.isZ80(props);
		boolean isROMX = (H89.isZ180(props) || H89.isNorbyFeat(props, "ROMX"));
		gpioIO1 = gpio.getIO1Bit();
		gpioMEM1 = gpio.getMEM1Bit();
		String s = props.getProperty("monitor_rom");
		if (isROMX) {
			if (s != null) { // TODO: assume AT27C256 if this???
				try {
					mon = new EPROM(s, 0x0000);
				} catch (Exception ee) {
					ee.printStackTrace();
					System.exit(1);
				}
			} else {
				// only 1 model supported, for now...
				romBit |= gpioIO1;
				mon = new EEPROM(props, "AT28C256", 0, intr);
			}
			if (!H89.isZ180(props)) {
				romBit |= gpioMEM1;
			}
			romMask = monMask = mon.length();
			if (monMask != 0x8000) {
				System.err.format("ROMX wrong size: %d\n",
								monMask);
			}
			return;
		}
		if (s == null) {
			if (isH8) {
				if (isZ80) {
					s = "2732_444-140_PAM37_v12.rom";
				} else {
					s = "2732_444-70_XCON8.rom";
				}
			} else {
				s = "2732a_444_84b_mms_v12.rom";
			}
			System.err.format("No Monitor ROM specified, using %s\n", s);
		}
		try {
			mon = new EPROM(s, 0);
		} catch (Exception ee) {
			ee.printStackTrace();
			System.exit(1);
		}
		monMask = mon.length();
		if (monMask != 0x1000 && monMask != 0x0800) {
			System.err.format("Monitor ROM wrong size: %d\n", monMask);
		}
		s = props.getProperty("floppy_rom");
		if (s == null) {
			//System.err.format("No Floppy (H17) ROM specified, using 444-79\n");
			s = "2716a_444_79.rom";
		}
		try {
			h17 = new EPROM(s, 0x1800);
		} catch (Exception ee) {
			ee.printStackTrace();
			System.exit(1);
		}
		int n = h17.length();
		if (n != 0x0800) {
			System.err.format("Floppy ROM wrong size: %d\n", n);
		}
		romMask = 0x2000;
	}

	public void romGpp(int gpio) {
		// TODO: limit this to ROMX situations?
		// should only get called if ROMX...
		if ((gpio & gpioMEM1) == 0) {
			// need RAM window 0x1000-0x17ff
			romMask = 0x2000;
			monMask = 0x1000;
		} else {
			romMask = monMask = mon.length();
		}
		// never for h17 ROM...
		mon.writeEnable((gpio & gpioIO1) != 0);
	}
}
