// Copyright (c) 2018 Douglas Miller <durgadas311@gmail.com>

import java.util.Arrays;
import java.util.Vector;
import java.util.Properties;
import java.io.*;
import java.awt.Point;
import javax.swing.*;
import javax.swing.Timer;
import java.awt.event.*;
import java.awt.*;

public class H8CassetteDrive implements DiskController, ActionListener,
		SerialDevice, GenericRemovableDrive, GenericRemovableMedia {
	private String Name = "Cassette";

	File name = null;
	RandomAccessFile file = null;
	LED led;
	boolean ledState = false;
	boolean protect = false;
	JPopupMenu pu;
	JMenuItem mnt;
	JMenuItem pos;
	JPanel pnl;	// container of led

	public H8CassetteDrive(Properties props, LEDHandler lh, Interruptor intr) {
		String s;

		String pfx = "cass";
		led = lh.registerLED(getDriveName(), LED.Colors.AMBER);
		Container c = led.getParent();
		if (c instanceof JPanel) {
			pnl = (JPanel)c;
			pnl.setToolTipText("EMPTY");
		}
		s = props.getProperty(pfx + "_tape");
		if (s != null) {
			insertMedia(s.split("\\s"));
		}
		// Usurp LED menu...
		JMenuItem mi = new JMenuItem("Cassette", KeyEvent.VK_P);
		mi.addActionListener(this);
		lh.setMenuItem(Name, mi);
		// Our Cassette Tape pop-up menu...
		pu = new JPopupMenu("Cassette");
		mi = new JMenuItem("Mount Tape", KeyEvent.VK_M);
		mi.addActionListener(this);
		pu.add(mi);
		mi = new JMenuItem("Rewind Tape", KeyEvent.VK_W);
		mi.addActionListener(this);
		pu.add(mi);
	}

	public Vector<GenericRemovableDrive> getDiskDrives() {
		Vector<GenericRemovableDrive> drives = new Vector<GenericRemovableDrive>();
		drives.add(this);
		return drives;
	}

	public String getDriveName(int index) {
		String str = Name;
		return str;
	}

	public GenericRemovableDrive findDrive(String ident) {
		if (Name.equals(ident)) {
			return this;
		}
		return null;
	}

	public void destroy() {
		eject(null);
	}

	// GenericRemovableDrive...

	public String getDriveName() { return "Cassette"; }
	public String getMediaName() {
		if (name != null) {
			return name.getName();
		}
		return "";
	}
	public void insertMedia(String[] media) {
		if (media == null) {
			eject(null);
			return;
		}
		protect = false;
		for (int x = 1; x < media.length; ++x) {
			if (media[x].equals("rw")) {
				protect = false;
			} else if (media[x].equals("ro")) {
				protect = true;
			}
		}
		name = new File(media[0]);
		try {
			file = new RandomAccessFile(name, protect ? "r" : "rw");
		} catch (Exception ee) {
			name = null;
		}
		updateText();
	}
	public int getMediaSize() { return mediaSize(); }
	public boolean isRemovable() { return true; }
	public boolean isReady() { return (file != null); }

	// GenericRemovableMedia...
	public boolean checkWriteProtect() { return false; } // TODO:
	public int mediaSize() {
		if (file != null) try {
			return (int)file.length();
		} catch (Exception ee) {}
		return 0;
	}
	public void eject(String name) {
		if (file != null) try {
			file.close();
		} catch (Exception ee) {}
		file = null;
		name = null;
		protect = false;
		updateText();
	}

	private void updateText() {
		if (pnl != null) {
			if (name == null) {
				pnl.setToolTipText("EMPTY");
			} else try {
				long p = file.getFilePointer();
				if (p < file.length()) {
					pnl.setToolTipText(
						String.format("%s (%d)", name.getName(), p));
				} else {
					pnl.setToolTipText(
						String.format("%s (END)", name.getName()));
				}
			} catch (Exception ee) {}
		}
	}

	// SerialDevice...

	public void rewind() {
		if (file != null) try {
			file.seek(0);
		} catch (Exception ee) {}
		updateText();
	}

	public void write(int b) {
		if (protect) {
			return;
		}
		if (file != null) try {
			file.write(b);
		} catch (Exception ee) { }
		updateText();
	}

	public int read() {
		int val = 0;
		if (file != null) try {
			val = file.read();
		} catch (Exception ee) {}
		updateText();
		return val;
	}

	public int available() {
		if (file != null) try {
			return (int)(file.length() - file.getFilePointer());
		} catch (Exception ee) {}
		return 0;
	}

	// VirtualUART.GET_RTS = Player Control
	// VirtualUART.GET_TXE = Recorder Control
	public void modemChange(VirtualUART it, int mdm) {
		ledState = ((mdm & (VirtualUART.GET_RTS|VirtualUART.GET_TXE)) != 0);
		led.set(ledState);
		if (it == null) {
			return;
		}
		if ((mdm & VirtualUART.GET_TXE) != 0) {
			it.setModem(VirtualUART.SET_DSR | VirtualUART.SET_CTS);
		} else {
			it.setModem(0);
		}
	}
	public int dir() { return SerialDevice.DIR_BIDI; }
	////////////////////////

	public void actionPerformed(ActionEvent e) {
		if (!(e.getSource() instanceof JMenuItem)) {
			return;
		}
		if (e.getActionCommand().equals("menu")) {
			mnt = (JMenuItem)e.getSource();
			return;
		}
		JMenuItem mi = (JMenuItem)e.getSource();
		int k = mi.getMnemonic();
		if (k == KeyEvent.VK_P) {
			// set location close to the action...
			// 'led' is the only graphic object we know about...
			pu.show(led, 0, 0);
		} else if (k == KeyEvent.VK_M) {
			if (mnt != null) {
				mnt.doClick();
			}
		} else if (k == KeyEvent.VK_W) {
			rewind();
		}
	}

	// IODevice needed by DiskController, but we're not an IODevice...
	public void reset() {}
	public int getBaseAddress() { return 0; }
	public int getNumPorts() { return -1; } // allow addDiskDevice() only
	public int in(int port) { return 0; }
	public void out(int port, int value) {}

	public String getDeviceName() { return Name; }

	public String dumpDebug() {
		long pos = 0;
		if (file != null) try {
			pos = file.getFilePointer();
		} catch (Exception ee) {}
		String ret = String.format("Cass=%s pos=%d\n",
			getMediaName(), pos);
		return ret;
	}
}
