// Copyright (c) 2018 Douglas Miller <durgadas311@gmail.com>

import java.util.Arrays;
import java.util.Properties;
import java.io.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.border.*;
import javax.swing.Timer;
import javax.sound.sampled.*;

public class H8FrontPanel implements GppListener, FPListener, KeyListener,
		ActionListener, MouseListener, WindowListener, Runnable {
	private Font tiny;
	private Font lesstiny;
	private Color wdw = new Color(70, 0, 0);
	private JFrame frame;
	JPanel panel;
	javax.swing.Timer timer;

	private H8Fpp gpp;
	private H89 sys;
	private Interruptor intr;
	private int src;

	long last;
	JLabel disp;
	char[] str;
	int[] digs;
	int[] xlat = new int[] {
		0,1,2, 4,5,6, 8,9,10
	};
	private LED pwr;
	private LED mon;
	private LED ion;
	private LED run;
	Clip _beep;
	JButton[] btns;
	// Default keypad labels to PAM-37 meanings
	String[] btx = new String[] {
		"0", "1", "2", "3", "4", "5", "6", "7",
		"8", "9", "A", "B", "C", "D", "E", "F"
	};
	String[] btl = new String[] {
		"BOOT H17 170 SP",
		"PRI H47 174 AF",
		"SEC H67 270 BC",
		"RADIX H37 274 DE",
		"GO HL",
		"IN IX",
		"OUT IY",
		"SI IR",
		"LOAD AF'",
		"DUMP BC'",
		"+ NEXT DE'",
		"- LAST HL'",
		"* CANCEL",
		"/ ALTER RST/\u2191",
		"# MEM RTM/\u2191",
		". REG PC"
	};
	int kpad;
	GridBagLayout gb;
	GridBagConstraints gc;

	public H8FrontPanel(Properties props, String pfx,
					H89 cmdr, H8Fpp gpio, Interruptor intr) {
		sys = cmdr;
		gpp = gpio;
		this.intr = intr;
		last = 0;
		tiny = new Font("Sans-serif", Font.PLAIN, 8);
		lesstiny = new Font("Sans-serif", Font.PLAIN, 10);
		digs = new int[9];
		str = new char[11];
		kpad = 0;
		btns = new JButton[16];
		Border lb = BorderFactory.createBevelBorder(BevelBorder.RAISED);
		Color bg = new Color(50, 50, 50);
		src = intr.registerINT(1);
		String s = props.getProperty("h8_keypad");
		if (s != null) {
			String[] ss = s.split(";");
			for (int x = 0; x < btl.length; ++x) {
				if (x < ss.length && ss[x].length() > 0) {
					if (ss[x].indexOf(":") > 0) {
						String[] sss = ss[x].split(":");
						btx[x] = sss[0];
						btl[x] = sss[1];
					} else {
						btl[x] = ss[x];
					}
				} else {
					btl[x] = null;
				}
			}
		}
		for (int x = 0; x < 16; ++x) {
			btns[x] = new JButton();
			btns[x].setPreferredSize(new Dimension(50, 50));
			btns[x].setBackground(bg);
			btns[x].setForeground(Color.white);
			btns[x].setBorder(lb);
			btns[x].setFocusPainted(false);
			btns[x].setPressedIcon(null);
			btns[x].addMouseListener(this);
			btns[x].setMnemonic(x + 0x1000);
			btns[x].setText(btx[x]);
			if (btl[x] != null) {
				btns[x].setToolTipText(btl[x]);
			}
		}
		pwr = new RoundLED(LED.Colors.RED);
		mon = new RoundLED(LED.Colors.RED);
		ion = new RoundLED(LED.Colors.RED);
		run = new RoundLED(LED.Colors.RED);
		pwr.set(true);
		setupBeep(props);

		panel = new JPanel();
		gb = new GridBagLayout();
		panel.setLayout(gb);
		panel.setOpaque(false);
		gc = new GridBagConstraints();
		gc.fill = GridBagConstraints.NONE;
		gc.gridx = 0;
		gc.gridy = 0;
		gc.weightx = 0;
		gc.weighty = 0;
		gc.gridwidth = 1;
		gc.gridheight = 1;
		gc.anchor = GridBagConstraints.CENTER;

		JPanel pan;
		pan = new JPanel();
		pan.setPreferredSize(new Dimension(10, 10));
		pan.setOpaque(false);
		gb.setConstraints(pan, gc);
		panel.add(pan);
		gc.gridx = 1;
		gc.gridy = 1;
		setDisplay();
		++gc.gridy;
		setNamePlate();
		--gc.gridy;
		gc.gridx += 3;
		setKeypad();
		++gc.gridx;
		gc.gridy += 2;
		pan = new JPanel();
		pan.setPreferredSize(new Dimension(10, 10));
		pan.setOpaque(false);
		gb.setConstraints(pan, gc);
		panel.add(pan);
		clearDisplay();

		frame = new JFrame("H8 Front Panel");
		//frame.getContentPane().setBackground(new Color(200,200,50));
		frame.add(panel);
		frame.pack();
		// TODO: what to do on window close...
		//frame.addWindowListener(this);
		frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		frame.addKeyListener(this);
		frame.setFocusable(true);
		frame.setVisible(true);
		gpp.addGppListener(this);
		gpp.setFPListener(this);
		timer = new javax.swing.Timer(50, this);
		timer.start();
	}

	public int interestedBits() { return 0x000080; }
	public void gppNewValue(int val) {
		if ((val & 0x080) != 0) {
			beepOff();
		} else {
			beepOn();
		}
	}

	public void update() {
		run.set(sys.isRunning());
		ion.set(sys.intEnabled());
	}

	public void timeout(int val) {
		mon.set((val & 0x020) != 0);
		int segs = ((val >> 8) & 0xff) ^ 0xff;
		int dig = (val & 0x0f);
		if (dig == 0) {
			return;
		}
		--dig;
		if (dig >= 9) {
			return;
		}
		// Detect when refreshing stops...
		last = System.nanoTime();
		if (digs[dig] == segs) {
			return;
		}
		digs[dig] = segs;
		if (segs < 32) {
			segs += 256;
		}
		str[xlat[dig]] = (char)segs;
		disp.setText(new String(str));
		disp.repaint();
	}

	private void clearDisplay() {
		Arrays.fill(digs, 0);
		Arrays.fill(str, (char)256);
		str[3] = '\002';
		str[7] = '\007';
		disp.setText(new String(str));
		disp.repaint();
	}

	private void setupBeep(Properties props) {
		String beep_wav = "h19beep.wav";
		String s = props.getProperty("h8_beep");
		if (s != null) {
			if (s.length() == 0) {
				_beep = null;
				return;
			}
			beep_wav = s;
		}
		try {
			AudioInputStream wav =
				AudioSystem.getAudioInputStream(
					new BufferedInputStream(
						this.getClass().getResourceAsStream(beep_wav)));
			AudioFormat format = wav.getFormat();
			DataLine.Info info = new DataLine.Info(Clip.class, format);
			_beep = (Clip)AudioSystem.getLine(info);
			_beep.open(wav);
			//_beep.setLoopPoints(0, loop); // ???
		} catch (Exception e) {
			e.printStackTrace();
			_beep = null;
			return;
		}
		int volume = 50;
		s = props.getProperty("h8_beep_volume");
		if (s != null) {
			volume = Integer.valueOf(s);
			if (volume < 0) volume = 0;
			if (volume > 100) volume = 100;
		}
		FloatControl vol = null;
		if (_beep.isControlSupported(FloatControl.Type.MASTER_GAIN)) {
			vol = (FloatControl)_beep.getControl(FloatControl.Type.MASTER_GAIN);
		} else if (_beep.isControlSupported(FloatControl.Type.VOLUME)) {
			vol = (FloatControl)_beep.getControl(FloatControl.Type.VOLUME);
		}
		if (vol != null) {
			float min = vol.getMinimum();
			float max = vol.getMaximum();
			float gain = (float)(min + ((max - min) * (volume / 100.0)));
			vol.setValue(gain);
		}
	}

	private void setNamePlate() {
		JPanel pan = new JPanel();
		pan.setPreferredSize(new Dimension(50, 10));
		pan.setOpaque(false);
		gb.setConstraints(pan, gc);
		panel.add(pan);
		++gc.gridx;
		Border lb = BorderFactory.createBevelBorder(BevelBorder.RAISED);
		JLabel pn = new JLabel();
		pn.setPreferredSize(new Dimension(300, 45));
		pn.setOpaque(true);
		pn.setBorder(lb);
		pn.setBackground(new Color(100,100,100));
		ImageIcon img = new ImageIcon(getClass().getResource("h8_nameplate.png"));
		pn.setIcon(img);
		gc.anchor = GridBagConstraints.SOUTHWEST;
		gb.setConstraints(pn, gc);
		panel.add(pn);
		gc.anchor = GridBagConstraints.CENTER;
		--gc.gridx;
	}

	private void setDisplay() {
		Font font = null;
		String f = "FND500.ttf";
		float fz = 45f;
			try {
				File ff = new File(f);
				java.io.InputStream ttf;
				if (ff.exists()) {
					ttf = new FileInputStream(ff);
				} else {
					ttf = H8FrontPanel.class.getResourceAsStream(f);
				}
				if (ttf != null) {
					font = Font.createFont(Font.TRUETYPE_FONT, ttf);
					font = font.deriveFont(fz);
				}
			} catch (Exception ee) {
				font = null;
System.err.format("No H8 FP font: %s\n", f);
			}
		JPanel pnl = new JPanel();
		pnl.setBackground(wdw);
		pnl.setOpaque(true);
		GridBagLayout _gb = new GridBagLayout();
		pnl.setLayout(_gb);
		GridBagConstraints _gc = new GridBagConstraints();
		_gc.fill = GridBagConstraints.NONE;
		_gc.gridx = 0;
		_gc.gridy = 0;
		_gc.weightx = 0;
		_gc.weighty = 0;
		_gc.gridwidth = 1;
		_gc.gridheight = 1;
		_gc.anchor = GridBagConstraints.CENTER;
		JPanel pn = new JPanel();
		pn.setPreferredSize(new Dimension(5, 5));
		pn.setOpaque(false);
		_gb.setConstraints(pn, _gc);
		pnl.add(pn);
		++_gc.gridx;
		++_gc.gridy;
		setColumn(ion, "ION", _gb, _gc, pnl);
		setColumn(mon, "MON", _gb, _gc, pnl);
		setColumn(run, "RUN", _gb, _gc, pnl);
		setColumn(pwr, "PWR", _gb, _gc, pnl);
		_gc.gridy -= 4;
		++_gc.gridx;	// skip LED
		++_gc.gridx;	// skip label
		_gc.gridheight = 3;
		disp = new JLabel();
		disp.setForeground(Color.red);
		disp.setBackground(wdw);
		disp.setOpaque(true);
		disp.setPreferredSize(new Dimension(450, 50));
		if (font != null) {
			disp.setFont(font);
		}
		_gb.setConstraints(disp, _gc);
		pnl.add(disp);
		_gc.gridheight = 1;
		_gc.gridy += 3;
		JLabel lab = new JLabel();
		lab.setForeground(Color.white);
		lab.setOpaque(false);
		lab.setFont(lesstiny);
		lab.setText("\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015 ADDRESS \u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015\u2015" +
				"\u2003\u2003\u2003\u2003" +
				"\u2015\u2015 DATA REGISTER \u2015\u2015");
		_gb.setConstraints(lab, _gc);
		pnl.add(lab);
		++_gc.gridx;
		++_gc.gridy;
		pn = new JPanel();
		pn.setPreferredSize(new Dimension(5, 5));
		pn.setOpaque(false);
		_gb.setConstraints(pn, _gc);
		pnl.add(pn);
		// Now add display unit to main panel...
		gc.gridwidth = 2;
		gc.anchor = GridBagConstraints.NORTH;
		gb.setConstraints(pnl, gc);
		panel.add(pnl);
		gc.anchor = GridBagConstraints.CENTER;
		gc.gridwidth = 1;
		gc.gridx += 2;
		JPanel pan = new JPanel();
		pan.setPreferredSize(new Dimension(50, 50));
		pan.setOpaque(false);
		gb.setConstraints(pan, gc);
		panel.add(pan);
		gc.gridx -= 2;
	}

	private void setKeypad() {
		JPanel pn = new JPanel();
		GridBagLayout _gb = new GridBagLayout();
		pn.setLayout(_gb);
		GridBagConstraints _gc = new GridBagConstraints();
		_gc.fill = GridBagConstraints.NONE;
		_gc.gridx = 0;
		_gc.gridy = 0;
		_gc.weightx = 0;
		_gc.weighty = 0;
		_gc.gridwidth = 1;
		_gc.gridheight = 1;
		_gc.anchor = GridBagConstraints.CENTER;
		_gb.setConstraints(btns[7], _gc);
		pn.add(btns[7]);
		++_gc.gridx;
		_gb.setConstraints(btns[8], _gc);
		pn.add(btns[8]);
		++_gc.gridx;
		_gb.setConstraints(btns[9], _gc);
		pn.add(btns[9]);
		++_gc.gridx;
		_gb.setConstraints(btns[10], _gc);
		pn.add(btns[10]);
		_gc.gridx = 0;
		++_gc.gridy;
		_gb.setConstraints(btns[4], _gc);
		pn.add(btns[4]);
		++_gc.gridx;
		_gb.setConstraints(btns[5], _gc);
		pn.add(btns[5]);
		++_gc.gridx;
		_gb.setConstraints(btns[6], _gc);
		pn.add(btns[6]);
		++_gc.gridx;
		_gb.setConstraints(btns[11], _gc);
		pn.add(btns[11]);
		_gc.gridx = 0;
		++_gc.gridy;
		_gb.setConstraints(btns[1], _gc);
		pn.add(btns[1]);
		++_gc.gridx;
		_gb.setConstraints(btns[2], _gc);
		pn.add(btns[2]);
		++_gc.gridx;
		_gb.setConstraints(btns[3], _gc);
		pn.add(btns[3]);
		++_gc.gridx;
		_gb.setConstraints(btns[12], _gc);
		pn.add(btns[12]);
		_gc.gridx = 0;
		++_gc.gridy;
		_gb.setConstraints(btns[0], _gc);
		pn.add(btns[0]);
		++_gc.gridx;
		_gb.setConstraints(btns[15], _gc);
		pn.add(btns[15]);
		++_gc.gridx;
		_gb.setConstraints(btns[14], _gc);
		pn.add(btns[14]);
		++_gc.gridx;
		_gb.setConstraints(btns[13], _gc);
		pn.add(btns[13]);

		gc.gridheight = 2;
		gb.setConstraints(pn, gc);
		panel.add(pn);
		gc.gridheight = 1;
	}

	private void setColumn(LED led, String str,
			GridBagLayout gb, GridBagConstraints gc, JPanel panel) {
		JLabel lbl = new JLabel(" " + str);
		gb.setConstraints(led, gc);
		panel.add(led);
		++gc.gridx;
		lbl.setPreferredSize(new Dimension(50, 16));
		lbl.setForeground(Color.white);
		lbl.setOpaque(false);
		lbl.setFont(tiny);
		gb.setConstraints(lbl, gc);
		panel.add(lbl);
		--gc.gridx;
		++gc.gridy;
	}

	private void beepOn() {
		if (_beep != null) {
			_beep.setFramePosition(0);
			_beep.loop(Clip.LOOP_CONTINUOUSLY);
		}
	}

	private void beepOff() {
		if (_beep != null) {
			_beep.stop();
			_beep.flush();
			// or? _beep.drain();
		}
	}

	// Keypad bits:
	// 0: kp1 pressed
	// 1: kp1.0
	// 2: kp1.1
	// 3: kp1.2
	// 4: kp2 pressed
	// 5: kp2.0
	// 6: kp2.1
	// 7: kp2.2
	public void mouseClicked(MouseEvent e) {}
	public void mouseEntered(MouseEvent e) {}
	public void mouseExited(MouseEvent e) {}
	public void mousePressed(MouseEvent e) {
		JButton btn = (JButton)e.getSource();
		int mn = btn.getMnemonic();
		mn &= 0x0f;
		int m = e.getModifiers();
		// Does not handle cases where SHIFT is pressed after key.
		if ((m & InputEvent.SHIFT_MASK) != 0) {
			if (mn == 13) {	// [0]+[/] or [0]+[D] "RST"
				// system RESET
				Thread t = new Thread(this);
				t.start();
				return;
			}
			if (mn == 14) {	// [0]+[#] or [0]+[E] "RTM"
				intr.raiseINT(1, src);
				gpp.setKey(0xd1);
				return;
			}
			return;
		}
//System.err.format("Press %02x\n", mn);
		kpadOn(mn);
	}
	public void mouseReleased(MouseEvent e) {
		JButton btn = (JButton)e.getSource();
		int mn = btn.getMnemonic();
		mn &= 0x0f;
		kpadOff(mn);
		// TODO: if SHIFT is released before this,
		// need to still clear interrupt (not tracking
		// SHIFT release events).
		// int m = e.getModifiers();
		// if ((m & InputEvent.SHIFT_MASK) != 0 && mn == 14) {
		if (mn == 14) { // [0]+[#] or [0]+[E] "RTM" released
			intr.lowerINT(1, src);
			return;
		}
	}

	private void kpadOff(int mn) {
		if (mn < 8) {
			kpad &= 0xee;
		} else {
			kpad &= 0xef;
		}
		gpp.setKey(kpad);
	}

	private void kpadOn(int mn) {
		if (mn < 8) {
			kpad = 0x00;
			kpad |= ((mn & 0x07) << 1) | 1;
		} else {
			kpad = 0x00;
			kpad |= (((mn & 0x07) << 1) | 1) << 4;
		}
		gpp.setKey(kpad);
	}

	public void windowActivated(WindowEvent e) { }
	public void windowClosed(WindowEvent e) { }
	public void windowIconified(WindowEvent e) { }
	public void windowOpened(WindowEvent e) { }
	public void windowDeiconified(WindowEvent e) { }
	public void windowDeactivated(WindowEvent e) { }
	public void windowClosing(WindowEvent e) {
		// TODO: allow hiding of panel?
		// if (e.getWindow() == frame) {
		// 	frame.setVisible(false);
		// 	return;
		// }
	}

	public void run() {
		sys.sendCommand("reset");
	}

	public void actionPerformed(ActionEvent e) {
		if (e.getSource() == timer) {
			if (last > 0 && System.nanoTime() - last > 50000000) {
				clearDisplay();
				last = 0;
			}
			return;
		}
	}

	private int key2kpad(char c) {
		if (Character.isDigit(c)) {
			return c - '0';
		}
		if (c >= 'A' && c <= 'F') {
			return (c - 'A') + 10;
		}
		switch(c) {
		case '+':
			return 10;
		case '-':
			return 11;
		case '*':
			return 12;
		case '/':
			return 13;
		case '#':
			return 14;
		case '.':
			return 15;
		}
		return -1;
	}

	public void keyTyped(KeyEvent e) { }
	public void keyPressed(KeyEvent e) {
		char c = Character.toUpperCase(e.getKeyChar());
		int m = e.getModifiers();
		if ((m & InputEvent.ALT_MASK) != 0) {
			if (c == '/' || c == 'D') {
				// system RESET
				Thread t = new Thread(this);
				t.start();
			}
			return;
		}
		int mn = key2kpad(c);
		if (mn < 0) {
			return;
		}
		kpadOn(mn);
	}
	public void keyReleased(KeyEvent e) {
		char c = Character.toUpperCase(e.getKeyChar());
		int m = e.getModifiers();
		if ((m & InputEvent.ALT_MASK) != 0) {
			return;
		}
		int mn = key2kpad(c);
		if (mn < 0) {
			return;
		}
		kpadOff(mn);
	}
}
