// Copyright (c) 2016 Douglas Miller <durgadas311@gmail.com>

import java.io.FileInputStream;
import java.util.Properties;
import java.io.*;

public class H8Memory extends H89Roms implements Memory, GppListener {
	// One single bank of 48K, plus ROMs... But RAM is fixed at 2000-E000...
	private byte[] mem;
	private int top;
	private int bot;
	private int h8_gppOrg0;
	private boolean rom;

	public H8Memory(Properties props, int low, int high,
			GeneralPurposePort gpio, Interruptor intr) {
		super(props, gpio, intr);	// cannot be ROMX
		int memsiz = 4096 * high;
		if (memsiz <= 0) {
			memsiz = 4096;
		}
		bot = 4096 * low;
		top = memsiz;
		mem = new byte[top];	// ensure mem at 0x0000
		h8_gppOrg0 = gpio.getOrg0Bit();
		reset();
		gpio.addGppListener(this);
	}

	public int read(boolean rom, int bank, int address) {
		address &= 0xffff; // necessary?
		if (rom && address < romMask) {
			if (address < monMask) {
				return mon.read(address);
			}
			if (address >= 0x1800 && h17 != null) {
				return h17.read(address & 0x07ff);
			}
		} else if (address >= top || address < bot) {
			return 0;
		}
		return mem[address] & 0xff;
	}
	public int read(int address) {
		return read(rom, 0, address);
	}
	public void write(int address, int value) {
		address &= 0xffff; // necessary?
		if (address >= top || address < bot) {
			return;
		}
		mem[address] = (byte)value;
	}

	public void reset() {
		rom = true;
	}

	public void dumpCore(String file) {
		try {
			OutputStream core = new FileOutputStream(file);
			core.write(mem);
			core.close();
		} catch (Exception ee) {
			H89Operator.error(null, "Core Dump", ee.getMessage());
		}
	}

	public int interestedBits() {
		return h8_gppOrg0;
	}
	public void gppNewValue(int gpio) {
		rom = ((gpio & h8_gppOrg0) == 0);
	}
}
