// Copyright (c) 2016 Douglas Miller <durgadas311@gmail.com>

import java.io.FileInputStream;
import java.util.Properties;
import java.io.*;

public class MMS77311Memory extends H89Roms implements Memory, GppListener {
	GeneralPurposePort gpp;
	int gpioBit;
	int gpioORG0;
	// One single bank of 64K, plus ROMs...
	private byte[] mem;
	private boolean rom;

	public MMS77311Memory(Properties props, GeneralPurposePort gpio, Interruptor intr) {
		super(props, gpio, intr);
		gpp = gpio;
		// rely entirely on notifications for 'rom' value.
		gpioORG0 = gpp.getOrg0Bit();
		gpioBit = romBit | gpioORG0;
		gpp.addGppListener(this);
		mem = new byte[65536];
	}

	public int read(boolean rom, int bank, int address) {
		address &= 0xffff; // necessary?
		if (rom && address < romMask) {
			if (address < monMask || (h17 == null && address >= 0x1800)) {
				return mon.read(address);
			}
			if (address >= 0x1800 && h17 != null) {
				return h17.read(address - 0x1800);
			}
		}
		return mem[address & 0xffff] & 0xff;
	}
	public int read(int address) {
		return read(rom, 0, address);
	}
	public void write(int address, int value) {
		if (rom && address < monMask) mon.write(address, value);
		mem[address & 0xffff] = (byte)(value & 0xff);
	}

	public void reset() {}

	public void dumpCore(String file) {
		try {
			OutputStream core = new FileOutputStream(file);
			core.write(mem);
			core.close();
		} catch (Exception ee) {
			H89Operator.error(null, "Core Dump", ee.getMessage());
		}
	}

	public int interestedBits() { return gpioBit; }
	public void gppNewValue(int gpio) {
		rom = ((gpio & gpioORG0) == 0);
		if (romBit != 0) romGpp(gpio);
	}
}
