// Copyright (c) 2017 Douglas Miller <durgadas311@gmail.com>

import java.io.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.border.*;
import java.util.Arrays;
import java.util.Properties;

public class Z67IDESASIDrive extends GenericSASIDrive
		implements ActionListener, WindowListener, Runnable {

	static final int GIGSHF = 30;		// Power-2 for 1G
	static final int MAXSHF = 27;		// Power-2 for 128M
	static final int MAXDRV = (1 << MAXSHF);// Each "drive" is 128M
	static final int MAXCYL = 2048;		// Num Cyl for MAXDRV
	static final int MAXHDS = 8;		// Num Heads for MAXDRV

	// TODO: create JFrame for control panel... and hook into front panel...
	int sel1, sel10;	// current state of selector switches
	long sysoff;	// Base offset for next I/O...
	long base;	// altered by selector switches
	long special;
	long cf;
	int numChunks;

	JFrame frame;
	JCheckBox[] wp;
	JButton d1u;
	JButton d1d;
	JButton d10u;
	JButton d10d;
	JLabel d1w;
	JLabel d10w;
	LED pwr;
	LED led;
	private javax.swing.Timer timer;
	Color sw = new Color(40, 40, 40);
	Color sw_h = new Color(60, 60, 60);
	Color sw_s = new Color(0, 0, 0);
	Font sw_f = new Font("Sans Serif", Font.BOLD, 8);
	JTextField cp_src1;
	JTextField cp_src2;
	JTextField cp_dst1;
	JTextField cp_dst2;
	JPanel cp_pn;
	Object[] cp_btn;
	static final int OPTION_CANCEL = 0;
	static final int OPTION_YES = 1;
	Thread thread;
	int curAction;
	int srcCard;
	int srcChunk;
	int dstCard;
	int dstChunk;
	String curCmd;

	Z67IDESASIDrive(Properties props, String name,
			String ctrl, String dir, int cnum, int capacity) {
		super(null, INVALID, name, ctrl, dir, cnum, 256);
		String p;
		cf = (long)capacity << GIGSHF;
		special = cf;
		String title = String.format("%s Z67-IDE+ 2x%dG", name, capacity);
		p = props.getProperty("z67ide_jukebox");
		if (p != null && (p.equals("true") || p.equals("yes") ||
				p.equals("on") || p.equals("1"))) {
			special = cf - (1L << (GIGSHF - 1));
			title += " (Jukebox)";
		}
		numChunks = (int)(special / MAXDRV);
		pwr = new RectLED(LED.Colors.GREEN);
		pwr.set(true);
		led = new RectLED(LED.Colors.RED);
		timer = new javax.swing.Timer(500, this);
		sel1 = 0;
		sel10 = 0;
		base = 0;
		driveFd = new RandomAccessFile[2];
		driveMedia = new String[2];
		String prefix = ctrl.toLowerCase();
		// This is only for mounting image files...
		devCfg = new int[] { 16384, capacity * 8, 0, 0 };
		for (int x = 0; x < 2; ++x) {
			p = String.format("%s_disk%d.%d", prefix, cnum, x);
			p = props.getProperty(p);
			if (p == null) {
				p = String.format("%s-%d.HD%d", ctrl, cnum, x);
			}
			driveMedia[x] = p;
			driveFd[x] = getLUN(p, dir, x);
			if (mediaSsz != 256 || mediaSpt != 32) {
				System.err.format("Invalid media for Z67-IDE: %s\n", p);
				// TODO: close it?
				driveFd[x] = null;
				driveSecLen = 256;
				sectorsPerTrack = 32;
			}
		}
		reset();
		System.err.format("Z67-IDE with %dG CFx2 on %s-%d\n", capacity, ctrl, cnum);
		wp = new JCheckBox[2];
		wp[0] = new JCheckBox("HD0");
		wp[0].setVerticalTextPosition(SwingConstants.TOP);
		wp[0].setHorizontalTextPosition(SwingConstants.CENTER);
		wp[0].setFocusable(false);
		wp[0].setFocusPainted(false);
		wp[1] = new JCheckBox("HD1");
		wp[1].setVerticalTextPosition(SwingConstants.TOP);
		wp[1].setHorizontalTextPosition(SwingConstants.CENTER);
		wp[1].setFocusable(false);
		wp[1].setFocusPainted(false);
		d1u = new JButton("+");
		d1u.addActionListener(this);
		d1u.setMargin(new Insets(1, 1, 1, 1));
		d1u.setPreferredSize(new Dimension(20, 20));
		d1d = new JButton("\u2500");
		d1d.addActionListener(this);
		d1d.setMargin(new Insets(1, 1, 1, 1));
		d1d.setPreferredSize(new Dimension(20, 20));
		d10u = new JButton("+");
		d10u.addActionListener(this);
		d10u.setMargin(new Insets(1, 1, 1, 1));
		d10u.setPreferredSize(new Dimension(20, 20));
		d10d = new JButton("\u2500");
		d10d.addActionListener(this);
		d10d.setMargin(new Insets(1, 1, 1, 1));
		d10d.setPreferredSize(new Dimension(20, 20));
		d1w = new JLabel("0");
		d1w.setBackground(Color.white);
		d1w.setHorizontalAlignment(SwingConstants.CENTER);
		d1w.setPreferredSize(new Dimension(14, 20));
		d1w.setOpaque(true);
		d10w = new JLabel("0");
		d10w.setBackground(Color.white);
		d10w.setHorizontalAlignment(SwingConstants.CENTER);
		d10w.setPreferredSize(new Dimension(14, 20));
		d10w.setOpaque(true);
		frame = new JFrame(title);
		GridBagLayout gb = new GridBagLayout();
		GridBagConstraints gc = new GridBagConstraints();
		frame.setLayout(gb);
		gc.gridx = 0;
		gc.gridy = 0;
		gc.anchor = GridBagConstraints.CENTER;
		JPanel pn = new JPanel();
		pn.setPreferredSize(new Dimension(50, 10));
		gb.setConstraints(pn, gc);
		frame.add(pn);
		++gc.gridx;
		++gc.gridy;
		// top and left gaps set...
		gb.setConstraints(pwr, gc);
		frame.add(pwr);
		++gc.gridy;
		JLabel lb = new JLabel("Power");
		gb.setConstraints(lb, gc);
		frame.add(lb);
		--gc.gridy;
		++gc.gridx;
		pn = new JPanel();
		pn.setPreferredSize(new Dimension(20, 10));
		gb.setConstraints(pn, gc);
		frame.add(pn);
		++gc.gridx;

		gc.gridwidth = 2;
		lb = new JLabel("Write-Protect");
		gb.setConstraints(lb, gc);
		frame.add(lb);
		gc.gridwidth = 1;
		gc.gridheight = 2;
		gc.anchor = GridBagConstraints.NORTH;
		++gc.gridy;
		gb.setConstraints(wp[0], gc);
		frame.add(wp[0]);
		++gc.gridx;
		gb.setConstraints(wp[1], gc);
		frame.add(wp[1]);
		gc.gridheight = 1;
		--gc.gridy;
		++gc.gridx;
		gc.anchor = GridBagConstraints.CENTER;

		pn = new JPanel();
		pn.setPreferredSize(new Dimension(20, 20));
		gb.setConstraints(pn, gc);
		frame.add(pn);
		++gc.gridx;

		gc.gridheight = 3;
		pn = makeThmbSw(d10u, d10d, d10w);
		gb.setConstraints(pn, gc);
		frame.add(pn);
		++gc.gridx;

		pn = makeThmbSw(d1u, d1d, d1w);
		gb.setConstraints(pn, gc);
		frame.add(pn);
		++gc.gridx;
		gc.gridheight = 1;

		pn = new JPanel();
		pn.setPreferredSize(new Dimension(20, 10));
		gb.setConstraints(pn, gc);
		frame.add(pn);
		++gc.gridx;
		gb.setConstraints(led, gc);
		frame.add(led);
		++gc.gridy;
		lb = new JLabel("Activity");
		gb.setConstraints(lb, gc);
		frame.add(lb);
		gc.gridy += 3;
		++gc.gridx;
		pn = new JPanel();
		pn.setPreferredSize(new Dimension(50, 10));
		gb.setConstraints(pn, gc);
		frame.add(pn);

		JMenuBar mb = new JMenuBar();
		JMenu mu;
		JMenuItem mi;
		mu = new JMenu("Util");
		mi = new JMenuItem("Cross-Copy", KeyEvent.VK_C);
		mi.addActionListener(this);
		mu.add(mi);
		mi = new JMenuItem("Replicate HD0", KeyEvent.VK_R);
		mi.addActionListener(this);
		mu.add(mi);
		mi = new JMenuItem("Replicate HD1", KeyEvent.VK_V);
		mi.addActionListener(this);
		mu.add(mi);
		mi = new JMenuItem("Image HD0 to HD1", KeyEvent.VK_I);
		mi.addActionListener(this);
		mu.add(mi);
		mi = new JMenuItem("Image HD1 to HD0", KeyEvent.VK_B);
		mi.addActionListener(this);
		mu.add(mi);
		mb.add(mu);
		frame.setJMenuBar(mb);

		cp_pn = new JPanel();
		cp_btn = new Object[2];
		cp_btn[OPTION_YES] = "Accept";
		cp_btn[OPTION_CANCEL] = "Cancel";
		cp_src1 = new JTextField();
		cp_src1.setPreferredSize(new Dimension(40, 20));
		cp_src2 = new JTextField();
		cp_src2.setPreferredSize(new Dimension(40, 20));
		cp_dst1 = new JTextField();
		cp_dst1.setPreferredSize(new Dimension(40, 20));
		cp_dst2 = new JTextField();
		cp_dst2.setPreferredSize(new Dimension(40, 20));
		cp_pn.setLayout(new BoxLayout(cp_pn, BoxLayout.Y_AXIS));
		pn = new JPanel();
		pn.add(new JLabel("Source drive, chunk:"));
		pn.add(cp_src1);
		pn.add(cp_src2);
		cp_pn.add(pn);
		pn = new JPanel();
		pn.add(new JLabel("Destination drive, chunk:"));
		pn.add(cp_dst1);
		pn.add(cp_dst2);
		cp_pn.add(pn);

		frame.pack();
		frame.addWindowListener(this);
	}

	private JPanel makeThmbSw(JButton up, JButton dn, JLabel wn) {
		JPanel pn = new JPanel();
		GridBagLayout gb = new GridBagLayout();
		pn.setLayout(gb);
		pn.setPreferredSize(new Dimension(40, 80));
		pn.setBackground(sw);
		pn.setBorder(BorderFactory.createBevelBorder(BevelBorder.RAISED,
				sw_h, sw_s));
		up.setBackground(sw);
		up.setPreferredSize(new Dimension(15, 15));
		up.setFocusable(false);
		up.setFocusPainted(false);
		up.setForeground(Color.white);
		up.setFont(sw_f);
		up.setBorder(BorderFactory.createBevelBorder(BevelBorder.RAISED,
				sw_h, sw_s));
		dn.setBackground(sw);
		dn.setBorder(BorderFactory.createBevelBorder(BevelBorder.RAISED,
				sw_h, sw_s));
		dn.setPreferredSize(new Dimension(15, 15));
		dn.setFocusable(false);
		dn.setFocusPainted(false);
		dn.setForeground(Color.white);
		dn.setFont(sw_f);
		wn.setBorder(BorderFactory.createBevelBorder(BevelBorder.LOWERED,
				sw_h, sw_s));
		wn.setPreferredSize(new Dimension(15, 20));
		GridBagConstraints gc = new GridBagConstraints();
		gc.insets.top = 2;
		gc.insets.bottom = 2;
		gc.insets.left = 2;
		gc.insets.right = 2;
		gc.gridx = 0;
		gc.gridy = 0;
		gc.anchor = GridBagConstraints.CENTER;
		gb.setConstraints(dn, gc);
		pn.add(dn);
		++gc.gridy;
		gb.setConstraints(wn, gc);
		pn.add(wn);
		++gc.gridy;
		gb.setConstraints(up, gc);
		pn.add(up);
		return pn;
	}

	private void reset() {
		// Default for chunks ("systems")... changed by OS
		devCfg[0] = MAXCYL;
		devCfg[1] = MAXHDS;
		devCfg[2] = 0;
		devCfg[3] = 0;
		capacity = devCfg[0] * devCfg[1] * sectorsPerTrack * driveSecLen;
	}

	public void register(LEDHandler lh) {
		if (lh != null) {
			String p = getDriveName();
			JMenuItem mi = new JMenuItem(p);
			mi.addActionListener(this);
			lh.setMenuItem(p, mi);
		}
	}

	@Override
	void resetSASI(SASIDriveController bus) {
		reset();
		super.resetSASI(bus);
	}

	@Override
	void seek(long off) throws Exception {
		if (sysoff < 0) {
			throw new IOException("Invalid Selector Switch");
		}
		// Current Z67IDE f/w just ORs in selector switches...
		driveFd[curLun].seek(sysoff | off);
	}

	@Override
	void processCmd(SASIDriveController bus) {
		curLun &= 1; // simulate Z67-IDE LUN enumeration
		sysoff = base;
		switch(cmdBuf[0]) {
		case cmd_FormatDrive_c:
		case cmd_FormatTrack_c:
		case cmd_FormatBadTrack_c:
			// Format commands don't actually do anything...
		case cmd_Write_c:
		case cmd_WriteLong_c:
			if (wp[curLun].isSelected()) {
				startError(bus, 0x03);	// write fault, close enough
				ack(bus);
				return;
			}
			// FALLTHROUGH
		case cmd_Read_c:
		case cmd_ReadLong_c:
			if (special < cf && cmdBuf[5] == 0x18) {
				sysoff = special;
			}
			break;
		}
		led.set(true);
		timer.removeActionListener(this);
		timer.addActionListener(this);
		timer.restart();
		super.processCmd(bus);
	}

	@Override
	void processDCB(SASIDriveController bus) {
		if (cmdBuf[0] != cmd_InitDriveChar_c) {
			// will this hang driver (if it ever happens)?
			return;
		}
		// TODO: enforce both drives being the same... ???
		int cyl = ((dcbBuf[0] & 0xff) << 8) | (dcbBuf[1] & 0xff);
		int hds = dcbBuf[2] & 0xff;
		int rwc = ((dcbBuf[3] & 0xff) << 8) | (dcbBuf[4] & 0xff);
		int wpc = ((dcbBuf[5] & 0xff) << 8) | (dcbBuf[6] & 0xff);
		// Don't care about ECC length = dcbBuf[7] & 0x0f;
		long cap = (long)cyl * hds * sectorsPerTrack * driveSecLen;
		devCfg[0] = cyl;
		devCfg[1] = hds;
		devCfg[2] = rwc;
		devCfg[3] = wpc;
		capacity = cap;
		if (cap > MAXDRV) {
			System.err.format("Host drive characteristics > 128M:" +
				" %02x %02x %02x %02x %02x %02x %02x %02x\n",
				dcbBuf[0], dcbBuf[1], dcbBuf[2], dcbBuf[3],
				dcbBuf[4], dcbBuf[5], dcbBuf[6], dcbBuf[7]);
			// TODO: return error? accept anyway? set to 128M?
			// Current Z67IDE f/w just forges ahead...
		}
		startStatus(bus);
		ack(bus);
	}

	private boolean copyChunk(int srcCard, int srcChunk, int dstCard, int dstChunk) {
		byte[] buf = new byte[64 * 1024];
		long srcOff = srcChunk << MAXSHF;
		long dstOff = dstChunk << MAXSHF;
		int num = MAXDRV / buf.length;
		// TODO: check to ensure system is RESET?
		led.set(true);
		try {
			// Need to seek each time, src and dst might be same file.
			for (int x = 0; x < num; ++x) {
				driveFd[srcCard].seek(srcOff);
				driveFd[srcCard].read(buf);
				driveFd[dstCard].seek(dstOff);
				driveFd[dstCard].write(buf);
				srcOff += buf.length;
				dstOff += buf.length;
			}
		} catch (Exception ee) {
			return false;
		} finally {
			led.set(false); // Assuming it was OFF to start...
		}
		return true;
	}

	private void copy() {
		int res = JOptionPane.showOptionDialog(frame, cp_pn,
			"Cross-Copy", JOptionPane.YES_NO_OPTION,
			JOptionPane.QUESTION_MESSAGE, null,
			cp_btn, cp_btn[OPTION_YES]);
		frame.requestFocus();
		if (res != OPTION_YES) {
			return;
		}
		if (cp_src1.getText().isEmpty() ||
			cp_src2.getText().isEmpty() ||
			cp_dst1.getText().isEmpty() ||
			cp_dst2.getText().isEmpty()) {
			PopupFactory.warning(frame, "Cross-Copy",
				"Must specify all values");
			return;
		}
		try {
			srcCard = Integer.valueOf(cp_src1.getText());
			srcChunk = Integer.valueOf(cp_src2.getText());
			dstCard = Integer.valueOf(cp_dst1.getText());
			dstChunk = Integer.valueOf(cp_dst2.getText());
		} catch (Exception ee) {
			PopupFactory.warning(frame, "Cross-Copy",
				"Invalid number");
			return;
		}
		if (srcCard < 0 || srcCard > 1 ||
			srcChunk < 0 || srcChunk >= numChunks ||
			dstCard < 0 || dstCard > 1 ||
			dstChunk < 0 || dstChunk >= numChunks) {
			PopupFactory.warning(frame, "Cross-Copy",
				"Invalid value");
			return;
		}
		if (srcCard == dstCard && srcChunk == dstChunk) {
			PopupFactory.warning(frame, "Cross-Copy",
				"Copy to same chunk");
			return;
		}
		curAction = KeyEvent.VK_C;
		thread = new Thread(this);
		thread.start();
	}

	private void doCopy() {
		String du = String.format("HD%d.%d to HD%d.%d",
				srcCard, srcChunk, dstCard, dstChunk);
		String cmd = "Copy " + du;
		int res = PopupFactory.confirm(frame, cmd,
			String.format("Overwrite HD%d chunk %d?", dstCard, dstChunk));
		if (res != JOptionPane.YES_OPTION) {
			return;
		}
		if (!copyChunk(srcCard, srcChunk, dstCard, dstChunk)) {
			PopupFactory.warning(frame, cmd, "Failed to copy " + du);
			return;
		}
		PopupFactory.inform(frame, cmd, "Copied " + du);
	}

	private void replicate(int card, int chunk) {
		curCmd = String.format("Replicate HD%d.%d", card, chunk);
		int res = PopupFactory.confirm(frame, curCmd,
			String.format("Overwrite HD%d chunks %d...?", card, chunk + 1));
		if (res != JOptionPane.YES_OPTION) {
			return;
		}
		srcCard = dstCard = card;
		srcChunk = chunk;
		dstChunk = chunk + 1;
		curAction = KeyEvent.VK_R;
		thread = new Thread(this);
		thread.start();
	}

	private void doReplicate() {
		String did = "";
		for (int x = dstChunk; x < numChunks; ++x) {
			String du = String.format("HD%d.%d to %d<BR>\n",
				srcCard, srcChunk, x);
			if (!copyChunk(srcCard, srcChunk, dstCard, x)) {
				PopupFactory.warning(frame, curCmd, "<HTML>" + did +
					"Failed to replicate " + du + "</HTML>");
				return;
			}
			did += "Replicated " + du;
		}
		PopupFactory.inform(frame, curCmd, "<HTML>" + did + "</HTML>");
	}

	private void image(int src, int dst) {
		curCmd = String.format("Image HD%d to HD%d", src, dst);
		int res = PopupFactory.confirm(frame, curCmd,
			String.format("Overwrite HD%d?", dst));
		if (res != JOptionPane.YES_OPTION) {
			return;
		}
		curAction = KeyEvent.VK_I;
		srcCard = src;
		srcChunk = 0;
		dstCard = dst;
		dstChunk = 0;
		thread = new Thread(this);
		thread.start();
	}

	private void doImage() {
		String did = "";
		for (int x = 0; x < numChunks; ++x) {
			String du = String.format("HD%d.%d to HD%d.%d<BR>\n",
				srcCard, x, dstCard, x);
			if (!copyChunk(srcCard, x, dstCard, x)) {
				PopupFactory.warning(frame, curCmd, "<HTML>" + did +
					"Failed to image " + du + "</HTML>");
				return;
			}
			did += "Imaged " + du;
		}
		PopupFactory.inform(frame, curCmd, "<HTML>" + did + "</HTML>");
	}

	public void actionPerformed(ActionEvent e) {
		if (e.getSource() == timer) {
			timer.removeActionListener(this);
			led.set(base < 0);
			return;
		}
		if (e.getSource() instanceof JButton) {
			JButton btn = (JButton)e.getSource();
			if (btn == d1u) {
				sel1 += 1;
				if (sel1 > 9) sel1 = 0;
			} else if (btn == d1d) {
				sel1 -= 1;
				if (sel1 < 0) sel1 = 9;
			} else if (btn == d10u) {
				sel10 += 1;
				if (sel10 > 9) sel10 = 0;
			} else if (btn == d10d) {
				sel10 -= 1;
				if (sel10 < 0) sel10 = 9;
			}
			d1w.setText(String.valueOf(sel1));
			d10w.setText(String.valueOf(sel10));
			base = (long)(sel10 * 10 + sel1) << MAXSHF;
			if (base + MAXDRV > special) {
				base = -1;
			}
			led.set(base < 0);
//System.err.format("SEL %d %d = %016x\n", sel10, sel1, base);
			return;
		}
		if (!(e.getSource() instanceof JMenuItem)) {
			return;
		}
		if (thread != null) {
			PopupFactory.warning(frame, "Util",
				"Operation in progress");
			return;
		}
		JMenuItem mi = (JMenuItem)e.getSource();
		int k = mi.getMnemonic();
		if (k == KeyEvent.VK_C) {
			// Cross-Copy a single chunk
			copy();
		} else if (k == KeyEvent.VK_R) {
			// Replicate HD0
			replicate(0, sel10 * 10 + sel1);
		} else if (k == KeyEvent.VK_V) {
			// Replicate HD1
			replicate(1, sel10 * 10 + sel1);
		} else if (k == KeyEvent.VK_I) {
			// Image HD0 to HD1
			image(0, 1);
		} else if (k == KeyEvent.VK_B) {
			// Image HD1 to HD0
			image(1, 0);
		}
		frame.setVisible(true);	// why?
	}
	public void windowActivated(WindowEvent e) { }
	public void windowClosed(WindowEvent e) { }
	public void windowIconified(WindowEvent e) { }
	public void windowOpened(WindowEvent e) { }
	public void windowDeiconified(WindowEvent e) { }
	public void windowDeactivated(WindowEvent e) { }
	public void windowClosing(WindowEvent e) {
		if (e.getWindow() == frame) {
			frame.setVisible(false);
			return;
		}
	}

	public void run() {
		if (curAction == KeyEvent.VK_I) {
			doImage();
		} else if (curAction == KeyEvent.VK_R) {
			doReplicate();
		} else if (curAction == KeyEvent.VK_C) {
			doCopy();
		}
		curAction = 0;
		thread = null;
	}
}
